unit BmpRes;

{ Load a bitmap from a resource. Unlike LoadBitmap, this one
  correctly loads the bitmap's palette. Also, this function
  creates and returns a TBitmap object.

  Tempest Software
  11 Nov 1995
}

interface

uses Classes, WinTypes, Graphics;

type
  { Stream for reading a resource. }
  TResourceStream = class(THandleStream)
  private
    fStartPos: LongInt;
    fEndPos: LongInt;
  public
    constructor Create(Instance: THandle; ResName, ResType: PChar);
    destructor Destroy; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
  end;

function LoadBitmapResource(Instance: THandle; Resource: PChar): TBitmap;

implementation

uses SysUtils, WinProcs;

type
  TBitmapStream = class(TResourceStream)
  private
    fPosition: LongInt;
  public
    constructor Create(Instance: THandle; ResName: PChar);
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    function Read(var Buffer; Count: LongInt): LongInt; override;
    function Write(const Buffer; Count: LongInt): LongInt; override;
  end;

{ Open a file to access the resources. The file position is automatically
  set to the start of the resource, but the TResourceStream object makes it
  appear that the resource starts at position zero, and that the size of
  the stream is the size of the resource. }
constructor TResourceStream.Create(Instance: THandle; ResName, ResType: PChar);
var
  ResInfo: THandle;
  Handle: Integer;
begin
  ResInfo := FindResource(Instance, ResName, ResType);
  if ResInfo = 0 then
    raise EResNotFound.Create('Cannot find resource');

  Handle := AccessResource(Instance, ResInfo);
  if Handle < 0 then
    raise EResNotFound.Create('Cannot access resource');

  inherited Create(Handle);
  fStartPos := inherited Seek(0, soFromCurrent);
  fEndPos := fStartPos + SizeOfResource(Instance, ResInfo);
end;

{ Close the file handle when we destroy the stream. }
destructor TResourceStream.Destroy;
begin
  if Handle >= 0 then
    FileClose(Handle);
  inherited Destroy;
end;

{ Create the illusion that the stream starts at zero and has the
  size of the resource. }
function TResourceStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
  soFromBeginning:
    Result := inherited Seek(fStartPos + Offset, Origin) - fStartPos;
  soFromCurrent:
    Result := inherited Seek(Offset, Origin) - fStartPos;
  soFromEnd:
    Result := inherited Seek(fEndPos + Offset, soFromBeginning) - fStartPos;
  else
    Result := -1;
  end;
end;

{ Load a TBitmap from a resource, creating a palette as needed. We use
  TBitmap.LoadFromStream to do the work for us, but we need to add
  a file header to the resource. We could choose to do this by making
  a copy of the bitmap into a temporary stream, but why make a redundant
  copy of what might be a large bitmap? Instead, we use a TBitmapStream
  that maintains a fake file header. }

{ Position 0 in the resource stream is position SizeOf(TBitmapFileHeader)
  in the bitmap stream. }
const
  BmpStartPos = SizeOf(TBitmapFileHeader);

constructor TBitmapStream.Create(Instance: THandle; ResName: PChar);
begin
  inherited Create(Instance, ResName, Rt_Bitmap)
end;

{ If we seek into the fake area (the first BmpStartPos bytes), then
  seek to the beginning of the stream. Otherwise, seek to the correct
  part of the stream, taking into account the BmpStartPos offset. }
function TBitmapStream.Seek(Offset: Longint; Origin: Word): Longint;
begin
  case Origin of
  soFromBeginning:
    Result := Offset;
  soFromCurrent:
    Result := fPosition + Offset;
  soFromEnd:
    Result := inherited Seek(0, soFromEnd) + Offset + BmpStartPos;
  else
    begin
      Result := -1;
      Exit;
    end;
  end;
  fPosition := Result;
  if fPosition > BmpStartPos then
    inherited Seek(fPosition - BmpStartPos, soFromBeginning)
  else
    inherited Seek(0, soFromBeginning);
end;

{ Read from the stream, possibly reading the bitmap file header.
  Note that this code assumes the file header does not cross a segment
  boundary. To make this code general purpose, we would need to
  handle such a possibility. }
function TBitmapStream.Read(var Buffer; Count: LongInt): LongInt;
var
  FileHdr: TBitmapFileHeader;
  BufStart: PChar;
  NBytes: Integer;
begin
  BufStart := @Buffer;
  if Position > BmpStartPos then
    NBytes := 0
  else
  begin
    FillChar(FileHdr, SizeOf(FileHdr), 0);
    FileHdr.bfType := (Ord('M') shl 8) or Ord('B');
    FileHdr.bfSize := Size;
    { We really should set bfOffBits, but that requires examining the
      resource. TBitmap.LoadFromStream does not use bfOffBits, so we
      don't bother to set it. If you ever want to use TBitmapStream
      in a more general use, then you may need to set bfOffBits. }
    NBytes := BmpStartPos - Position;
    Move(FileHdr, BufStart^, NBytes);
    Inc(BufStart, NBytes);
    Dec(Count, NBytes);
    Inc(fPosition, NBytes);
  end;
  Result := NBytes + inherited Read(BufStart^, Count);
  Inc(fPosition, Count);
end;

{ Ignore any attempt to write the file header }
function TBitmapStream.Write(const Buffer; Count: LongInt): LongInt;
var
  NBytes: Integer;
  BufStart: PChar;
begin
  BufStart := @Buffer;
  if Position > BmpStartPos then
    NBytes := 0
  else
  begin
    NBytes := BmpStartPos - Position;
    Inc(BufStart, NBytes);
    Dec(Count, NBytes);
    Inc(fPosition, NBytes);
  end;
  Result := NBytes + inherited Write(BufStart^, Count);
  Inc(fPosition, Count);
end;

{ Load a bitmap resource, with its palette, and create a TBitmap object. }
function LoadBitmapResource(Instance: THandle; Resource: PChar): TBitmap;
var
  ResStream: TBitmapStream;
begin
  ResStream := TBitmapStream.Create(Instance, Resource);
  try
    Result := TBitmap.Create;
    try
      Result.LoadFromStream(ResStream);
    except
      Result.Free;
      raise;
    end;
  finally
    ResStream.Free;
  end;
end;

end.
{--=>revision-history<=--}
{1 BMPRES.PAS 19-Nov-95,16:49:40,`lisch' CHANGE=INITIAL                       }
{     Initial version.                                                        }
{--=>revision-history<=--}
