 unit BjTypes;

interface

uses SysUtils, Classes;

const
   DECKCARDS = 52;         // Number of cards per deck (duh!)

type
   TCard = record
      Face        : byte;  // 1-4 = Heard, Diamond, Club, Spade
      Num         : byte;  // A=1,2,3,4,5,6,7,8,9,10,J=11,Q=12,K=13
      BjVal       : byte;  // A=11,J=10,Q=10,K=10, else 2=2, etc
      Junk        : byte;  // PlaceHolder
   end;

   TCardArray = array[0..999999] of TCard;
   
   TPlayerAction = (paHit, paStand, paDouble, paSplit);

   TDeck = class
   private
      FNumDecks   : integer;
      FCards      : ^TCardArray;
      FPenet      : double;
      CurCard     : integer;              // the current card in the deck
      CutAfter    : integer;              // which card has the "Yellow" cut card
      procedure   SetDecks(i: integer);
      procedure   ReleaseMem;
      function    GetNeedShuffle:boolean;
      procedure   SetPenet(p: double);
   public
      Count       : integer;
      constructor Create;
      destructor  Destroy; override;
      procedure   Shuffle;
      function    DealCard: TCard;
      property    NumDecks:integer read FNumDecks write SetDecks default 0;
      property    Penetration:double read FPenet write SetPenet;
      property    NeedShuffle:boolean read GetNeedShuffle;
      function    TrueCount: double;
   end;

   THandOptions = set of (hoCanSplit, hoCanDouble, hoCanHit);

   THand = class
   private
      FCards      : array[0..10] of TCard;   // Can't have more than this, can it?
      FNumCards   : integer;
      HaveAce     : boolean;
      LowTot      : integer;
      WasSplit    : boolean;
   public
      Wager       : double;
      Options     : THandOptions;
      constructor Create;
      procedure   InitHand;
      procedure   AddCard(c: TCard);
      function    Total: integer;
      function    IsSoft: boolean;
      function    IsBlackJack: boolean;
      function    ShowCard: TCard;
      function    HavePair: boolean;
      property    NumCards:integer read FNumCards;
      function    Out: string;
      procedure   TakeHit(Deck: TDeck);
      procedure   Split(NewHand: THand; Deck: TDeck);
   end;

   TPlayer = class
   public
      Cash        : double;
      NumHands    : integer;
      Hands       : array[0..9] of THand;
      constructor Create;
      destructor  Destroy; override;
      procedure   InitHands(Wager: double);
      procedure   PlayHands(DealerCard: TCard; Deck: TDeck);
      function    EvaluateAllHands(DealerHand: THand): double;
      function    AllHandsBusted: boolean;
   end;


implementation

uses DebugFm;

// *** TDeck ***

constructor TDeck.Create;
begin
   FNumDecks := 0;
   FPenet    := 0.80;
   FCards    := nil;
end;

destructor TDeck.Destroy;
begin
   ReleaseMem;
end;

procedure TDeck.SetDecks(i: integer);
const
   Lookup : array[0..12] of integer =
      (1,2,3,4,5,6,7,8,9,10,10,10,10);
var
   u, v : integer;
begin
   ReleaseMem;
   FNumDecks := i;
   GetMem(FCards, FNumDecks * DECKCARDS * sizeof(TCard));

   // Initialize the cards
   for u := 0 to i-1 do
      for v := 0 to DECKCARDS-1 do
         with FCards^[u * DECKCARDS + v] do begin
            Face  := v DIV 13 + 1;
            Num   := v MOD 13 + 1;
            BjVal := Lookup[Num-1];
            Junk  := 0;
         end;

   CurCard  := FNumDecks * DECKCARDS;           // so don't accidentally deal from unshuffled deck
   CutAfter := Trunc(FNumDecks * DECKCARDS * Penetration);
end;

procedure TDeck.SetPenet(p: double);
begin
   FPenet := p;
   CutAfter := Trunc(FNumDecks * DECKCARDS * Penetration);
end;

procedure TDeck.ReleaseMem;
begin
   if FCards<>nil then
      FreeMem(FCards, FNumDecks * DECKCARDS * sizeof(TCard));
   FCards := nil;
end;

procedure TDeck.Shuffle;
var
   i, j : integer;
   t    : TCard;
begin
   for i := 0 to FNumDecks * DECKCARDS - 1 do begin
      j := Random(FNumDecks * DECKCARDS);
      t := FCards^[i];
      FCards^[i] := FCards^[j];
      FCards^[j] := t;
   end;
   CurCard := 0;
   Count   := 0;
end;

function TDeck.GetNeedShuffle:boolean;
begin
   Result := CurCard > CutAfter;
end;

function TDeck.DealCard: TCard;
const
   CountVal : array[1..10] of integer =
      (-1, 1, 1, 1, 1, 1, 0, 0, 0, -1);   // Ace & Ten are -1, 2-6 are +1
begin
   if CurCard >= FNumDecks * DECKCARDS then
      raise Exception.Create('Dealt past the end of the deck!');
   Result := FCards^[CurCard];
   Count := Count + CountVal[Result.BjVal];
   CurCard := CurCard + 1;
end;

function TDeck.TrueCount: double;
var
   DecksRemain : double;
begin
   DecksRemain := (FNumDecks * DECKCARDS - CurCard) / DECKCARDS;
   Result := Count / DecksRemain;
end;


// *** THand ***

constructor THand.Create;
begin
   InitHand;
end;

procedure THand.InitHand;
begin
   FNumCards := 0;
   HaveAce   := False;
   LowTot    := 0;
   Options   := [hoCanSplit, hoCanDouble, hoCanHit];
   WasSplit  := False;
end;

procedure THand.AddCard(c: TCard);
begin
   FCards[FNumCards] := c;
   FNumCards := FNumCards + 1;
   if c.Num=1 then HaveAce := True;
   LowTot := LowTot + c.BjVal;
end;

function THand.Total: integer;
begin
   Result := LowTot;
   if HaveAce and (LowTot < 12) then Result := LowTot + 10;
end;

function THand.IsSoft: boolean;
begin
   Result := HaveAce and (LowTot < 12);
end;

function THand.IsBlackJack: boolean;
begin
   Result := (FNumCards = 2) and (Total = 21) and (not WasSplit);
end;

function THand.ShowCard: TCard;
begin
   if FNumCards < 1 then
      raise Exception.Create('Hand as no cards to show!');
   Result := FCards[0];
end;

function THand.HavePair: boolean;
begin
   {if FNumCards <> 2 then
      raise Exception.Create('Hand does not have 2 cards!');}
   Result := (FNumCards = 2) and (FCards[0].Num = FCards[1].Num);
end;

function CardOut(c: TCard): string;
const
   CardStr : array[1..13] of string =
      ('A', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'J', 'Q', 'K');
   SuitStr : array[1..4] of string =
      ('h', 'd', 's', 'c');
begin
   Result := CardStr[c.Num] + SuitStr[c.Face];
end;

function THand.Out: string;
var
   i : integer;
   t : string;
begin
   t := '';
   for i := 0 to FNumCards-1 do
      t := t + CardOut(FCards[i]) + ' ';
   t := t + '(' + IntToStr(Total) + ')';
   Result := t;
end;

procedure THand.TakeHit(Deck: TDeck);
var
   nc : TCard;
begin
   if not (hoCanHit in Options) then
      raise Exception.Create('Hit when hand was not allowed');
   nc := Deck.DealCard;
   AddCard(nc);
{$IFDEF Debug}
   Debug('Hit Card = ' + CardOut(nc) + '   ('+IntToStr(Total)+')');
{$ENDIF}
end;

// NewHand is the new hand that gets created after a split
procedure THand.Split(NewHand: THand; Deck: TDeck);
begin
   // Set this hand to just one card
   FNumCards := 1;
   LowTot := FCards[0].BjVal;
   WasSplit := True;
   Options := Options - [hoCanSplit];

   NewHand.InitHand;
   NewHand.Wager := Wager;
   NewHand.Options := Options;
   NewHand.AddCard(FCards[1]);      // Give the new hand the second card
   NewHand.WasSplit := True;
   NewHand.Options := NewHand.Options - [hoCanSplit];

   // If Split Aces, Modify the options of both hands so they can't split again
   if (FCards[0].Num=1) then begin
      Options := Options - [hoCanHit];
      NewHand.Options := NewHand.Options - [hoCanHit];
   end;

   // Give Each hand another card
   AddCard(Deck.DealCard);
   NewHand.AddCard(Deck.DealCard);

   {$IFDEF Debug}
   Debug('*************************************************** Splitting!');
   Debug('Hand 1 = ' + Out);
   Debug('Hand 2 = ' + NewHand.Out);
   {$ENDIF}
end;


// *** TPlayer ***

constructor TPlayer.Create;
var
   i : integer;
begin
   NumHands := 0;
   for i := 0 to 9 do
      Hands[i] := THand.Create;
end;

destructor TPlayer.Destroy;
var
   i : integer;
begin
   for i := 0 to 9 do
      Hands[i].Free;
end;

procedure TPlayer.InitHands(Wager: double);
begin
   NumHands := 1;
   Hands[0].InitHand;
   Hands[0].Wager := Wager;
end;

function Trans(c: char): TPlayerAction;
begin
   case c of
      'h'  : Result := paHit;
      'd'  : Result := paDouble;
      'p'  : Result := paSplit;
      's'  : Result := paStand;
   else
      Result := paStand;
   end;
end;

function BasicStrategy(PlayerHand: THand; DealerCard: TCard): TPlayerAction;
const
   HardDoubleTab : array[9..11, 1..10] of char =
    //  A    2    3    4    5    6    7    8    9   10
     (('h', 'h', 'd', 'd', 'd', 'd', 'h', 'h', 'h', 'h'),      // 9
      ('h', 'd', 'd', 'd', 'd', 'd', 'd', 'd', 'd', 'h'),      // 10
      ('d', 'd', 'd', 'd', 'd', 'd', 'd', 'd', 'd', 'd'));     // 11
   SplitTab : array[1..10, 1..10] of char =
    //  A    2    3    4    5    6    7    8    9   10
     (('p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p'),      // A
      ('h', 'h', 'h', 'p', 'p', 'p', 'p', 'h', 'h', 'h'),      // 2
      ('h', 'h', 'h', 'p', 'p', 'p', 'p', 'h', 'h', 'h'),      // 3
      ('h', 'h', 'h', 'h', 'h', 'h', 'h', 'h', 'h', 'h'),      // 4
      ('d', 'd', 'd', 'd', 'd', 'd', 'd', 'd', 'h', 'h'),      // 5
      ('h', 'p', 'p', 'p', 'p', 'p', 'h', 'h', 'h', 'h'),      // 6
      ('h', 'p', 'p', 'p', 'p', 'p', 'p', 'h', 'h', 'h'),      // 7
      ('p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p'),      // 8
      ('s', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 's'),      // 9
      ('s', 's', 's', 's', 's', 's', 's', 's', 's', 's'));     // 10
   SoftTab : array[12..21, 1..10] of char =
    //  A    2    3    4    5    6    7    8    9   10
     (('p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p', 'p'),      // A-A 12
      ('h', 'h', 'h', 'h', 'd', 'd', 'h', 'h', 'h', 'h'),      // A-2
      ('h', 'h', 'h', 'h', 'd', 'd', 'h', 'h', 'h', 'h'),      // A-3
      ('h', 'h', 'h', 'd', 'd', 'd', 'h', 'h', 'h', 'h'),      // A-4
      ('h', 'h', 'h', 'd', 'd', 'd', 'h', 'h', 'h', 'h'),      // A-5
      ('h', 'h', 'd', 'd', 'd', 'd', 'h', 'h', 'h', 'h'),      // A-6
      ('h', 's', 'd', 'd', 'd', 'd', 's', 's', 'h', 'h'),      // A-7
      ('s', 's', 's', 's', 's', 's', 's', 's', 's', 's'),      // A-8
      ('s', 's', 's', 's', 's', 's', 's', 's', 's', 's'),      // A-9
      ('s', 's', 's', 's', 's', 's', 's', 's', 's', 's'));     // A-10
   StiffTab : array[12..16, 1..10] of char =
    //  A    2    3    4    5    6    7    8    9   10
     (('h', 'h', 'h', 's', 's', 's', 'h', 'h', 'h', 'h'),      // 12
      ('h', 's', 's', 's', 's', 's', 'h', 'h', 'h', 'h'),      // 13
      ('h', 's', 's', 's', 's', 's', 'h', 'h', 'h', 'h'),      // 14
      ('h', 's', 's', 's', 's', 's', 'h', 'h', 'h', 'h'),      // 15
      ('h', 's', 's', 's', 's', 's', 'h', 'h', 'h', 'h'));     // 16
var
   tot  : integer;
begin
   tot  := PlayerHand.Total;

   // Hard Double Down strategy
   if PlayerHand.NumCards = 2 then
      if (tot >= 9) and (tot <=11) then begin
         Result := Trans(HardDoubleTab[tot, DealerCard.BjVal]);
         exit;
      end;

   // Splitting Strategy
   if (PlayerHand.HavePair) and (hoCanSplit in PlayerHand.Options) then begin
      Result := Trans(SplitTab[PlayerHand.ShowCard.BjVal, DealerCard.BjVal]);
      exit;
   end;

   // Soft Hand Strategy
   if PlayerHand.IsSoft then begin
      Result := Trans(SoftTab[tot, DealerCard.BjVal]);
      exit;
   end;

   // Hard Hand Strategy for Stiffs
   if tot < 12 then begin
      Result := paHit;
      exit;
   end;
   if tot > 16 then begin
      Result := paStand;
      exit;
   end;
   if (tot <= 16) and (tot >= 12) then begin
      Result := Trans(StiffTab[tot, DealerCard.BjVal]);
      exit;
   end;

   raise Exception.Create('Never should get here!');
end;

procedure TPlayer.PlayHands(DealerCard: TCard; Deck: TDeck);
label
   PlayHand, DonePlaying;
const
   BustCards : set of byte = [2,3,4,5,6];
var
   i      : integer;
   Action : TPlayerAction;
begin
   i := 0;
   while (i <= NumHands-1) do begin
      // Play out Hand[i]
PlayHand:
      // If can't hit it means you are done with this hand
      if not (hoCanHit in Hands[i].Options) then
         goto DonePlaying;

      Action := BasicStrategy(Hands[i], DealerCard);

      if Action = paStand then
         goto DonePlaying;

      // Split Aces or Eights
      if Action = paSplit then begin
         // split
         Hands[i].Split(Hands[NumHands], Deck);  // Give Split func the new hand to put cards into
         NumHands := NumHands + 1;
         goto PlayHand;
      end;

      if (Action = paDouble) and (Hands[i].NumCards = 2) then begin
         // double
         {$IFDEF Debug}
         Debug('*************************************************** Double!');
         {$ENDIF}
         Hands[i].TakeHit(Deck);
         Hands[i].Options := Hands[i].Options - [hoCanHit];
         Hands[i].Wager := Hands[i].Wager * 2;

         goto DonePlaying;
      end;

      if (Action = paHit) or (Action = paDouble) then begin
         Hands[i].TakeHit(Deck);
         goto PlayHand;
      end;

DonePlaying:
   i := i + 1;
   end; // i loop
end;

function EvaluateHand(PlayerHand, DealerHand: THand):double;
label
   Win, Tie, Lose;
var
   pt, dt : integer;
begin
   pt := PlayerHand.Total;
   dt := DealerHand.Total;

   // Check if the Player's Hand busted
   if pt > 21 then begin
      {$IFDEF Debug}
      Debug(Format('Player Busted! (%d)', [pt]));
      {$ENDIF}
      goto Lose;
   end;

   if DealerHand.IsBlackjack and (not PlayerHand.IsBlackjack) then
      goto Lose;

   if (PlayerHand.IsBlackjack) and (dt<>21) then begin
      {$IFDEF Debug}
      Debug(Format('BlackJack!  Win = $%1.0n', [PlayerHand.Wager * 3/2]));
      {$ENDIF}
      Result := PlayerHand.Wager * 3/2;
      exit;
   end;

   if dt > 21 then goto Win;
   if pt > dt then goto Win;
   if pt < dt then goto Lose;

   // Reached a tie
Tie:
   {$IFDEF Debug}
   Debug(Format('Push (%d=%d)', [pt,dt]));
   {$ENDIF}
   Result := 0;
   exit;

Win:
   {$IFDEF Debug}
   Debug(Format('Player Wins $%1.0n! (%d over %d)', [PlayerHand.Wager, pt, dt]));
   {$ENDIF}
   Result := PlayerHand.Wager;
   exit;

Lose:
   {$IFDEF Debug}
   Debug(Format('Player loses $%1.0n (%d is under %d)', [PlayerHand.Wager, pt, dt]));
   {$ENDIF}
   Result := -PlayerHand.Wager;
end;

// Retuns the net winnings for these hands
function TPlayer.EvaluateAllHands(DealerHand: THand): double;
var
   i   : integer;
   tot : double;
begin
   tot := 0;
   for i := 0 to NumHands-1 do
      tot := tot + EvaluateHand(Hands[i], DealerHand);

   Result := tot;
end;

function TPlayer.AllHandsBusted: boolean;
var
   AllBust : boolean;
   i       : integer;
begin
   AllBust := True;
   for i := 0 to NumHands-1 do
      if Hands[i].Total < 22 then begin
         AllBust := False;
         break;
      end;
   Result := AllBust;
end;


end.
