{**************************************************************************}
{*                   Business.pas for Delphi4, Delphi5                    *}
{*                   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~                    *}
{*               Version 1.10, 02/Oct/1999, 22:50:00 (MSK)                *}
{* Copyright(C) 1996-1999 by Atsioukovski Alexandre. All Rights Reserved  *}
{*  Portions Copyright(C) 1999 by United Software. All Rights Reserved    *}
{* Portions Copyright(C) 1998 by TurboPower Software. All Rights Reserved *}
{*                                                                        *}
{*                         E-Mail: united@lgg.ru                          *}
{*                       Business common routines                         *}
{**************************************************************************}

Unit Business;

Interface
Uses Classes, SysUtils, Math;

{String routines}
Function StrToFloatEx(Const S: String): Extended;
//Converts given string with not standard decimal separators to a floating-point value
Function FloatToStrEx(X: Extended; DigitsAfterPoint: Byte): String;
//Converts a floating point value to a string rounded to DigitsAfterPoints
Function FloatToCurrencyStr(X: Extended): String;
//Converts a floating point value to a currency string rounded to 2 digits after point
Function ReplaceAll(Const S, SearchStr, ReplaceStr: String): String;
//Replaces all SearchStr to ReplaceStr in a string
Function ReplaceRecursive(Const S, SearchStr, ReplaceStr: String): String;
//Replaces all SearchStr to ReplaceStr until SearchStr is not found in a string again
Function TrimDoubleSpaces(Const S: String): String;
//Trims a string of leading and trailing spaces and double spaces
Function TrimPointsRight(Const S: String): String;
//Trims a string of leading and trailing points
Function TrimExtended(Const S: String; DoTrimPoints: Boolean = False): String;
//Trims a string of leading and trailing spaces, double spaces and points
Function StringWOLastChar(Const S: String): String;
//Returns a string without last char
Function GetLeftDigits(S: String): String;
//Returns a string with all left digits of S
Function GetAfterLeftDigits(S: String): String;
//Returns a string with all chars after left digits of S
Function LeadingZeroStr(S: String; NumDigits: Integer = 2): String;
//Returns a string with leading zeros
Function GetUniqueStringList(Const StringList: TStringList): TStringList;
//Gets a unique string list

{Time routines}
Function CalcTotalMSecs(StartMSecs, CurrentMSecs: Comp;
                        Current, Total: Integer): Comp;
//Calculates how much time it will take to finish the operation
Function MSecsToStr(MSecs: Comp): String;
//Converts converts the number of milliseconds into a normal time string

{Arithmetic routines}
Function RoundEx(X: Extended; DigitsAfterPoint: Byte = 0): Extended;
//Returns the value of X rounded to the nearest whole number with DigitsAfterPoint
Function RoundToFactor(X, Factor: Extended): Extended;
//Returns the value of X rounded to the nearest Factor number
Function RoundUpToFactor(X, Factor: Extended): Extended;
//Returns the value of X rounded up toward nearest Factor number

{Statistical routines}
Function Forecast(X: Double; Const KnownX, KnownY: Array Of Double): Double;

{Data routines}
Function MergeArrays(Var A, B: Array Of Double): Boolean;
//Merges two arrays by using zero values as the empty cells
Function MinNonZeroValue(Data: Array Of Double; Var Index: Integer): Double;
//Returns smallest signed non-zero value in an array
Function SecondMinNonZeroValue(Data: Array Of Double; Var Index: Integer): Double;
//Returns the second smallest signed non-zero value in an array

{Money routines (in Russian language)}
Function IntToNumStr(Value: Integer): String;
//Returns number in words string
Function MoneyToNumStr(Value: Currency; RusCurr: Boolean = False): String;
//Returns money in words string

Implementation
Const
 cMinus: Char = '-';
 cPoint: Char = ',';
 cDecimalPoint: Char = '.';
 sSpace: String = ' ';
 sDoubleSpace: String = '  ';

{String routines}
Function ExchangeChars(S: String; FromChar, ToChar: Char): String;
Var
 I: Integer;
Begin
 Result := '';
 For I := 1 To Length(S) Do
  If S[I] = FromChar Then
   Result := Result + ToChar Else
   Result := Result + S[I];
End;

Function NormalizeDecimals(Const Value: String): String;
Begin
 Result := ExchangeChars(Value, cMinus, DecimalSeparator);
 Result := ExchangeChars(Result, cPoint, DecimalSeparator);
 Result := ExchangeChars(Result, cDecimalPoint, DecimalSeparator);
End;

Function StrToFloatEx(Const S: String): Extended;
Begin
 Try
  Result := StrToFloat(NormalizeDecimals(Trim(S)));
 Except
  Result := 0;
 End;
End;

Function FloatToStrEx(X: Extended; DigitsAfterPoint: Byte): String;
Begin
 Result := FloatToStr(RoundEx(X, DigitsAfterPoint));
End;

Function FloatToCurrencyStr(X: Extended): String;
Begin
 Result := Trim(Format('%8.2f', [X]));
End;

Function ReplaceAll(Const S, SearchStr, ReplaceStr: String): String;
Var
 I: Integer;
 Source: String;
Begin
 Source := S;
 Result := '';
 Repeat
  I := Pos(SearchStr, Source);
  If I > 0 Then
   Begin
    Result := Result + Copy(Source, 1, I - 1) + ReplaceStr;
    Source := Copy(Source, I + Length(SearchStr), MaxInt);
   End Else Result := Result + Source;
 Until I <= 0;
End;

Function ReplaceRecursive(Const S, SearchStr, ReplaceStr: String): String;
Var
 I: Integer;
Begin
 Result := S;
 Repeat
  I := Pos(SearchStr, Result);
  If I <> 0 Then Result := ReplaceAll(Result, SearchStr, ReplaceStr);
 Until I <= 0;
End;

Function TrimDoubleSpaces(Const S: String): String;
Begin
 Result := Trim(ReplaceRecursive(S, sDoubleSpace, sSpace));
End;

Function TrimPointsRight(Const S: String): String;
Var
 I: Integer;
Begin
 If S = '' Then Exit;
 I := Length(S);
 While (I > 0) And (S[I] = '.') Do Dec(I);
 Result := Copy(S, 1, I);
End;

Function TrimExtended(Const S: String; DoTrimPoints: Boolean = False): String;
Begin
 Result := S;
 If DoTrimPoints Then Result := TrimPointsRight(Trim(Result));
 Result := TrimDoubleSpaces(Result);
End;

Function StringWOLastChar(Const S: String): String;
Begin
 Result := Copy(S, 1, Length(S) - 1);
End;

Function GetLeftDigits(S: String): String;
Var
 I: Integer;
Begin
 Result := '';
 S := Trim(S);
 For I := 1 To Length(S) Do
  If S[I] In ['0'..'9'] Then
   Result := Result + S[I] Else
   Break;
End;

Function GetAfterLeftDigits(S: String): String;
Var
 I, P: Integer;
Begin
 S := Trim(S);
 If GetLeftDigits(S) = '' Then
  Begin
   Result := S;
   Exit;
  End;  
 P := 1;
 For I := 1 To Length(S) Do
  If Not (S[I] In ['0'..'9']) Then
   Begin
    P := I;
    Break;
   End;
 If P <> 1 Then
  Result := Copy(S, P, Length(S) - P + 1);
End;

Function LeadingZeroStr(S: String; NumDigits: Integer = 2): String;
Var
 Zeros: String;
 I: Integer;
Begin
 Result := S;
 If Length(Result) < NumDigits Then
  Begin
   Zeros := '';
   For I := 1 To (NumDigits - Length(Result)) Do Zeros := Zeros + '0';
   Result := Zeros + Result;
  End;
End;

Function GetUniqueStringList(Const StringList: TStringList): TStringList;
Var
 I: Integer;
 CurrentString: String;
Begin
 StringList.Sort;
 Result := TStringList.Create;
 Result.Clear;
 CurrentString := #255#255#255#255#255;
 For I := 0 To StringList.Count - 1 Do
  If ANSICompareText(StringList.Strings[I], CurrentString) <> 0 Then
   Begin
    If StringList.Strings[I] <> '' Then
     Result.Add(StringList.Strings[I]);
    CurrentString := StringList.Strings[I];
   End;
End;

{Time routines}
Function CalcTotalMSecs(StartMSecs, CurrentMSecs: Comp;
                        Current, Total: Integer): Comp;
Begin
 If Current <> 0 Then
  Result := Round(((CurrentMSecs - StartMSecs) * Total) / Current) Else
  Result := 0;
End;

Function MSecsToStr(MSecs: Comp): String;
Var
 DateTime: TDateTime;
Begin
 DateTime := TimeStampToDateTime(MSecsToTimeStamp(MSecs));
 Result := FormatDateTime('hhhh:mm:ss', DateTime);
End;

{Arithmetic routines}
Function RoundEx(X: Extended; DigitsAfterPoint: Byte = 0): Extended;
Begin
 Result := StrToFloat(Trim(Format('%20.' +
                      IntToStr(DigitsAfterPoint) + 'f', [X])));
End;

Function RoundToFactor(X, Factor: Extended): Extended;
Begin
 Result := Factor * Int((X + (Factor / 2)) / Factor);
End;

Function RoundUpToFactor(X, Factor: Extended): Extended;
Begin
 Result := Factor * Int(((X + Factor) + (Factor / 2)) / Factor);
 If (Result - X) >= Factor Then
  Result := Factor * Int((X + (Factor / 2)) / Factor);
End;

{Statistical routines}
Type
 TDoubleArray = Array[0..(MaxLongInt Div SizeOf(Double)) - 1] Of Double;

Type
 TStLinEst = Record
  B0, B1: Double; //Model coefficients
  seB0, seB1: Double; //Standard error of model coefficients
  R2: Double; //Coefficient of determination
  Sigma: Double; //Standard error of regression
  SSr, SSe: Double; //Elements for ANOVA table
  F0: Double; //F-statistic to test B1=0
  Df: Integer; //denominator degrees of freedom for F-statistic
End;

Procedure LinEst16(Const KnownY; Const KnownX; NData: Integer;
                   Var LF: TStLinEst; ErrorStats: Boolean);
Var
 I: Integer;
 SX, SY, XMean, YMean, SXX, SXY, SYY, X, Y: Extended;
Begin
 If (NData <= 2) Then Exit;
 {Compute basic sums}
 SX := 0.0;
 SY := 0.0;
 SXX := 0.0;
 SXY := 0.0;
 SYY := 0.0;
 For I := 0 To NData-1 Do
  Begin
   X := TDoubleArray(KnownX)[I];
   Y := TDoubleArray(KnownY)[I];
   SX := SX + X;
   SY := SY + Y;
   SXX := SXX + X * X;
   SYY := SYY + Y * Y;
   SXY := SXY + X * Y;
  End;
 XMean := SX / NData;
 YMean := SY / NData;
 SXX := SXX - NData * XMean * XMean;
 SYY := SYY - NData * YMean * YMean;
 SXY := SXY - NData * XMean * YMean;
 {Check for zero variance}
 If (SXX <= 0.0) Or (SYY <= 0.0) Then Exit;
 {Initialize returned parameters}
 FillChar(LF, SizeOf(LF), 0);
 {Regression coefficients}
 LF.B1 := SXY / SXX;
 LF.B0 := YMean - LF.B1 * XMean;
 {Error statistics}
 If (ErrorStats) Then
  Begin
   LF.SSr := LF.B1 * SXY;
   LF.SSe := SYY - LF.SSr;
   LF.R2 := LF.SSr / SYY;
   LF.Df := NData - 2;
   LF.Sigma := Sqrt(LF.SSe / LF.Df);
   If LF.SSe = 0.0 Then
    LF.F0 := 1.7E+308 Else //Pick an arbitrarily large number for perfect fit
    LF.F0 := (LF.SSr * LF.Df) / LF.SSe;
   LF.seB1 := LF.Sigma / Sqrt(SXX);
   LF.seB0 := LF.Sigma * Sqrt((1.0 / NData) + (XMean * XMean / SXX));
  End;
End;

Function Forecast16(X: Double; Const KnownY; Const KnownX; NData: Integer): Double;
Var
 LF: TStLinEst;
Begin
 LinEst16(KnownY, KnownX, NData, LF, False);
 Result := LF.B0 + LF.B1 * X;
End;

Function Forecast(X: Double; Const KnownX, KnownY: Array Of Double): Double;
Begin
 Result := Forecast16(X, KnownY, KnownX, High(KnownY) + 1);
End;

{Data routines}
Function MergeArrays(Var A, B: Array Of Double): Boolean;
Var
 I: Integer;
Begin
 Result := False;
 For I := Low(A) To High(A) Do If A[I] <> B[I] Then Result := True;
 If Result = False Then Exit;
 For I := Low(A) To High(A) Do
  Begin
   If (A[I] = 0) And (B[I] <> 0) Then A[I] := B[I];
   If (A[I] <> 0) And (B[I] = 0) Then B[I] := A[I];
  End;
End;

Function MinNonZeroValue(Data: Array Of Double; Var Index: Integer): Double;
Var
 I: Integer;
 AllZeros: Boolean;
Begin
 AllZeros := True;
 For I := Low(Data) To High(Data) Do
  If Data[I] = 0 Then Data[I] := MaxDouble Else AllZeros := False;
 If AllZeros Then
  Begin
   Result := 0;
   Exit;
  End;
 Result := Data[Low(Data)];
 For I := Low(Data) + 1 To High(Data) Do
  If Result > Data[I] Then
   Result := Data[I];
 For I := Low(Data) To High(Data) Do
  If Data[I] = Result Then
   Begin
    Index := I;
    Break;
   End;
End;

Function SecondMinNonZeroValue(Data: Array Of Double; Var Index: Integer): Double;
Var
 MValue: Double;
 MValueIndex: Integer;
Begin
 MValue := MinNonZeroValue(Data, MValueIndex);
 If MValue = 0 Then
  Begin
   Result := 0;
   Exit;
  End;
 Data[MValueIndex] := 0;
 Result := MinNonZeroValue(Data, Index);
End;

{Money routines (in Russian language)}
Const
 RusCurrencyMain: Array[0..2] Of String =
  ('', '', '');
 RusCurrencySub: Array[0..2] Of String =
  ('', '', '');
 USCurrencyMain: Array[0..2] Of String =
  ('', '', '');
 USCurrencySub: Array[0..2] Of String =
  ('', '', '');

Const
 FemaleOne: Array[0..8] Of String =
  ('', '', '', '', '', 
   '', '', '', '');
 MaleOne: Array[0..8] Of String =
  ('', '', '', '', '',
   '', '', '', '');
 FirstTen: Array[0..8] Of String =
  ('', '', '', '',
   '', '', '', '',
   '');
 Ten: Array[0..8] Of String =
  ('', '', '', '', '',
   '', '', '', '');
 Hundred: Array[0..8] Of String =
  ('', '', '', '', '',
   '', '', '', '');
 Thousand: Array[0..2] Of String =
  ('', '', '');
 Million: Array[0..2] Of String =
  ('', '', '');

Const
 Zero = '';

Function Triada(Value: Integer; MaleGender: Boolean): String;
Var
 Tmp: Integer;
Begin
 Tmp := Value Div 100;
 Value := Value Mod 100;
 If Tmp > 0 Then
  Begin
   Result := Hundred[Tmp - 1];
   If Value > 0 Then Result := Result + ' ';
  End Else Result := '';
 Tmp := Value Div 10;
 Value := Value Mod 10;
 If Tmp > 1 Then
  Begin
   Result := Result + Ten[Tmp - 1];
   Tmp := Value;
   If Tmp > 0 Then
    Begin
     Result := Result + ' ';
     If MaleGender Then
      Result := Result + MaleOne[Tmp - 1] Else
      Result := Result + FemaleOne[Tmp - 1];
    End;
  End Else
  Begin
   Tmp := Tmp * 10 + Value;
   Case Tmp Of
    1..9:
     Begin
      If MaleGender Then
       Result := Result + MaleOne[Tmp - 1] Else
       Result := Result + FemaleOne[Tmp - 1];
     End;
    10: Result := Result + Ten[0];
    11..19: Result := Result + FirstTen[Tmp - 11];
   End;
  End;
End;

Function IntegerCurrencyStr(Value: Integer; Curr: Array Of String): String;
Begin
 If Value Mod 100 In [11..19] Then
  Result := Curr[2] Else
  Case (Value Mod 10) Of
   0, 5..9: Result := Curr[2];
   1: Result := Curr[0];
   2..4: Result := Curr[1];
  End;
End;

Function IntegerGenderedToString(Value: Integer; MaleGender: Boolean): String;
Var
 Tmp: Integer;
Begin
 If Value = 0 Then
  Begin
   Result := Zero;
   Exit;
  End;
 Tmp := Value Div 1000000;
 Value := Value Mod 1000000;
 If Tmp > 0 Then
  Begin
   Result := IntegerGenderedToString(Tmp, True) + ' ' +
             IntegerCurrencyStr(Tmp, Million);
   If Value > 0 Then Result := Result + ' ';
  End Else Result := '';
 Tmp := Value Div 1000;
 Value := Value Mod 1000;
 If Tmp > 0 Then
  Begin
   Result := Result + IntegerGenderedToString(Tmp, False) + ' ' +
             IntegerCurrencyStr(Tmp, Thousand);
   If Value > 0 Then Result := Result + ' ';
  End;
 Tmp := Value;
 If Tmp > 0 Then Result := Result + Triada(Tmp, MaleGender);
End;

Function IntToNumStr(Value: Integer): String;
Begin
 Result := IntegerGenderedToString(Value, True);
End;

Function MoneyToNumStr(Value: Currency; RusCurr: Boolean = False): String;
Var
 Curr, CurrSub: Integer;
 Tmp, ResultSub: String;
Begin
 Curr := Trunc(Value);
 CurrSub := Round(Frac(Value) * 100);
 If CurrSub = 100 Then
  Begin
   CurrSub := 0;
   Inc(Curr);
  End;
//CurrencyMain string
 If Curr <> 0 Then
  Result := IntegerGenderedToString(Curr, True) Else //Rubles or dollars
  Result := '';
 If Result <> '' Then
  Begin
   If Not RusCurr Then
    Tmp := IntegerCurrencyStr(Curr, USCurrencyMain) Else
    Tmp := IntegerCurrencyStr(Curr, RusCurrencyMain);
   If Tmp <> '' Then Result := Result + ' ' + Tmp;
  End;
//CurrencySub string
 If CurrSub <> 0 Then
  Begin
   If Not RusCurr Then
    ResultSub := IntegerGenderedToString(CurrSub, True) Else //Cents
    ResultSub := IntegerGenderedToString(CurrSub, False); //Copecs
  End Else
  ResultSub := '';
 If ResultSub <> '' Then
  Begin
   If Not RusCurr Then
    Tmp := IntegerCurrencyStr(CurrSub, USCurrencySub) Else
    Tmp := IntegerCurrencyStr(CurrSub, RusCurrencySub);
   If Tmp <> '' Then ResultSub := ResultSub + ' ' + Tmp;
  End;
//CurrencyMain + CurrencySub
 If Result <> '' Then
  Begin
   If ResultSub <> '' Then
    Result := Result + ' ' + ResultSub;
  End Else Result := Result + ResultSub;
End;

End.
