{=============================================================================
  Unit     : ROLLUPS.PAS

  Author   : Rob McGregor, 73122.3125@compuserve.com

  Purpose  : This unit provides a customized Delphi TForm Object that
             contains methods for implementing a basic "Rollup" form
             class.

  Comments : Forms derived from this base class inherit all the rollup
             functionality herein...

  History  : Date      Reason
             --------  ---------------------------------------------------
             11-14-95  Created basic rollup class.
==============================================================================}

unit Rollups;

interface

uses
  SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
  Forms, Dialogs, Buttons, Menus, ExtCtrls;

const
  ROLL_AMOUNT   = 15;      { Sets the rollup / rolldown speed }
  SC_MOUSEMOVE  = $F012;   { Used for dragging the window from fake caption }

type
  TRollupForm = class(TForm)

  {** Protected Methods **}

  protected
    procedure RU_FormDeactivate (Sender: TObject; btnRollUp, btnRollDown,
      btnClose: TSpeedButton);

    procedure RU_FormMouseDown (Sender: TObject; X, Y: Integer);
    procedure RU_FormMouseMove (Sender: TObject; X, Y: Integer);
    procedure RU_FormMouseUp;

    procedure RU_FormPaint (Sender: TObject; btnRollUp,
      btnRollDown, btnClose: TSpeedButton);

    procedure RU_Init (Sender: TObject; btnRollUp, btnRollDown,
      btnClose, btnMenu: TSpeedbutton);

    procedure RU_RollDown (Sender: TObject; btnRollUp, btnRollDown,
      btnClose: TSpeedButton);

    procedure RU_RollUp (Sender: TObject; btnRollUp, btnRollDown,
      btnClose: TSpeedButton);

    procedure RU_ShowMenu (Sender: TObject; btnMenu: TSpeedButton;
      PopupMenu: TPopupMenu);

  private
    {** Private Variables **}

    rcTitleBar       : TRect;                 { Title bar rectangle      }
    nMaxClientHeight : integer;               { Max height for rollup    }
    bDragging        : boolean;               { State variable           }
    szCaption        : array[0..78] of char;  { The caption string       }
    nCaptionLength   : integer;               { The length of the string }

    {** Private Methods **}

    function  RU_CreateDCLogFont (dc: HDC; nHeight: integer;
                                  yStyle: byte): HFont;
    procedure RU_DrawTitleBar (btnRollUp, btnRollDown, btnClose: TSpeedButton);
    procedure RU_DrawWindowBorder;
    function  RU_IsPointInFakeCaption (x, y: integer): boolean;
  end;


implementation

{*****************************************************************************
 *                                                                           *
 *                            Protected Methods                              *
 *                                                                           *
 *****************************************************************************}

{-----------------------------------------------------------------------------
   Procedure : RU_FormDeactivate

   Purpose   : Redraws the fake title bar as inactive.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_FormDeactivate (Sender: TObject; btnRollUp,
                                         btnRollDown, btnClose: TSpeedButton);
begin
  RU_DrawTitleBar (btnRollUp, btnRollDown, btnClose);

end;  { RU_FormDeactivate }


{-----------------------------------------------------------------------------
   Procedure : RU_FormMouseDown

   Purpose   : Set the drag flag for moving the form.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_FormMouseDown (Sender: TObject; X, Y: Integer);
begin
  if (RU_IsPointInFakeCaption (x, y)) then
    bDragging := true;

end;  { RU_FormMouseDown }


{-----------------------------------------------------------------------------
   Procedure : RU_FormMouseMove

   Purpose   : Moves the form with the mouse if the drag flag is set.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_FormMouseMove (Sender: TObject; X, Y: Integer);
var
  fRet: real;

begin
  if (bDragging) then
  begin
    fRet := SendMessage (handle, WM_LBUTTONUP, 0, 0);
    fRet := SendMessage (handle, WM_SYSCOMMAND, SC_MOUSEMOVE, 0);
  end;
end;


{-----------------------------------------------------------------------------
   Procedure : RU_FormMouseUp

   Purpose   : Resets the drag flag.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_FormMouseUp;
begin
  bDragging := false;
end;


{-----------------------------------------------------------------------------
   Procedure : RU_FormPaint

   Purpose   : Redraws the fake title bar as active and redraws the 3d border.

   Comments  : Call this method from FormActivate as well as FormPaint.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_FormPaint (Sender: TObject; btnRollUp,
                                    btnRollDown, btnClose: TSpeedButton);
begin
  RU_DrawTitleBar (btnRollUp, btnRollDown, btnClose);
  RU_DrawWindowBorder;

end;  { RU_FormPaint }


{-----------------------------------------------------------------------------
   Procedure : RU_Init

   Purpose   : Initializes the data and controls for a rollup window.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_Init (Sender: TObject; btnRollUp, btnRollDown,
                               btnClose, btnMenu: TSpeedbutton);
begin
  nMaxClientHeight := self.clientheight;
  btnRollUp.left   := btnRollDown.left;

  { Set the rollup form's basic windows attributes }
  self.FormStyle   := fsStayOnTop;
  self.BorderStyle := bsNone;

  { Set the dimensions of the fake caption bar }
  with rcTitleBar do
  begin
    left   := 3;
    top    := 3;
    right  := clientwidth - 3;
    bottom := 20;
  end;

  { Move the titlebar buttons into position (no matter what the form width) }
  btnClose.left    := clientwidth - btnClose.width - 4;
  btnRollUp.left   := btnClose.left - btnRollup.width - 4;
  btnRollDown.left := btnClose.left - btnRollup.width - 4;
  btnMenu.left     := btnClose.left;

  btnClose.top    := 4;
  btnRollUp.top   := 4;
  btnRollDown.top := 4;
  btnMenu.top     := 25;

  { Set the fake titlebar caption }
  nCaptionLength := length (self.caption) + 2;  { We'll add 2 blank spaces at
                                                  the beginning of the string }
  StrPCopy (szCaption, '  ' + self.caption);

  { OK, a little Windows API magic... }
  RU_DrawTitleBar (btnRollUp, btnRollDown, btnClose);
  RU_DrawWindowBorder;

end;  { RU_Init }


{-----------------------------------------------------------------------------
   Procedure : RU_RollDown

   Purpose   : Rolls the form down to its max clientheight.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_RollDown (Sender: TObject; btnRollUp, btnRollDown,
                                   btnClose: TSpeedButton);
var
  nCH, i : integer;

begin
  nCH := clientheight;

  { Erase window border left after after a roll-up }
  self.canvas.pen.color := clBtnFace;
  self.canvas.moveto (2, rcTitleBar.bottom);
  self.canvas.lineto (clientwidth - 2, rcTitleBar.bottom);
  self.canvas.moveto (2, rcTitleBar.bottom + 1);
  self.canvas.lineto (clientwidth - 2, rcTitleBar.bottom + 1);

  self.canvas.moveto (2, rcTitleBar.bottom + 2);
  self.canvas.lineto (clientwidth - 2, rcTitleBar.bottom + 2);

  { The actual "rolldown" action... }
  while (clientheight < nMaxClientHeight) do
    clientheight := clientheight + ROLL_AMOUNT;

  btnRollUp.visible   := true;
  btnRollDown.visible := false;

  clientheight := nMaxClientHeight;
  RU_DrawTitleBar (btnRollUp, btnRollDown, btnClose);
  RU_DrawWindowBorder;

end;  { RU_RollDown }


{-----------------------------------------------------------------------------
   Procedure : RU_RollUp

   Purpose   : Rolls the form up into a floating title bar.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_RollUp (Sender: TObject; btnRollUp, btnRollDown,
                                 btnClose: TSpeedButton);
var
  nCH: integer;

begin
  nCH := clientheight;

  while (clientheight > btnRollDown.height + 8) do
    clientheight := clientheight - ROLL_AMOUNT;

  btnRollDown.visible := true;
  btnRollUp.visible   := false;

  clientheight := btnRollDown.height + 8;
  RU_DrawWindowBorder;

end;  { RU_RollUp }


{-----------------------------------------------------------------------------
   Procedure : RU_ShowMenu

   Purpose   : Invokes the popup menu at the specified location.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_ShowMenu (Sender: TObject; btnMenu: TSpeedButton; PopupMenu: TPopupMenu);
var
  x1, y1: integer;

begin
  btnMenu.refresh;
  x1 := self.left + btnMenu.left + btnMenu.width;
  y1 := self.top + btnMenu.top;
  PopupMenu.popup (x1, y1);

end;  { RU_ShowMenu }


{*****************************************************************************
 *                                                                           *
 *                            Private Methods                                *
 *                                                                           *
 *****************************************************************************}

{-----------------------------------------------------------------------------
   Procedure : RU_CreateDCLogFont

   Purpose   : Creates a logical font of the typestyle and size desired,
               selects it into the device context.

   Returns   : Handle to the new logical font.

   Comments  : The logical font must be destroyed using DeleteObject after
               use!!
 -----------------------------------------------------------------------------}
function TRollupForm.RU_CreateDCLogFont (dc: HDC; nHeight: integer;
                                         yStyle: byte): HFont;
var
  lf : TLogFont;
  hf : HFont;

begin
  { select a font and fill in the structure }
  with lf do
  begin
    lfHeight         := nHeight;
    lfWidth          := 0;
    lfEscapement     := 0;
    lfOrientation    := 0;
    lfWeight         := FW_THIN;
    lfItalic         := 0;
    lfUnderline      := 0;
    lfStrikeOut      := 0;
    lfCharSet        := 0;
    lfOutPrecision   := OUT_TT_ONLY_PRECIS;
    lfClipPrecision  := CLIP_TT_ALWAYS;
    lfQuality        := DEFAULT_QUALITY;
    lfPitchAndFamily := DEFAULT_PITCH or yStyle;
  end;

  { Create the font }
  hf := CreateFontIndirect (lf);
  SelectObject (dc, hf);

  result := hf;

end;  { RU_CreateDCLogFont }


{-----------------------------------------------------------------------------
   Procedure : RU_DrawTitleBar

   Purpose   : Draws a fake mini-caption bar on the top of the form.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_DrawTitleBar (btnRollUp, btnRollDown,
                                       btnClose: TSpeedButton);
var
  dc      : HDC;
  bActive : boolean;
  hf      : HFont;

begin
  with canvas do
  begin
    dc := GetDC (self.handle);

    { Is this rollup active? }
    bActive := (screen.ActiveForm = self);

    if (bActive) then
    begin
      brush.color := GetSysColor (COLOR_ACTIVECAPTION);
      SetTextColor (dc, GetSysColor (COLOR_CAPTIONTEXT));
    end
    else begin
      brush.color := GetSysColor (COLOR_INACTIVECAPTION);
      SetTextColor (dc, GetSysColor (COLOR_INACTIVECAPTIONTEXT));
    end;

    { Draw the fake caption bar }
    FillRect (rcTitleBar);

    { Draw the caption text }
    hf := RU_CreateDCLogFont (dc, 15, FF_SWISS);
    SetBkMode (dc, TRANSPARENT);
    DrawText (dc, szCaption, nCaptionLength, rcTitleBar, DT_LEFT);

    ReleaseDC (self.handle, dc);
    DeleteObject (hf);       { Kill the logical font }

    { Make sure the titlebar buttons aren't painted over by the caption! }
    btnClose.refresh;
    btnRollup.refresh;
    btnRolldown.refresh;
  end;
end;  { RU_DrawTitleBar }


{-----------------------------------------------------------------------------
   Procedure : RU_DrawWindowBorder

   Purpose   : Draws a 3d border around the window.
 -----------------------------------------------------------------------------}
procedure TRollupForm.RU_DrawWindowBorder;
var
  nCH, nCW: integer;

begin
  with canvas do
  begin
    nCW := clientwidth - 1;
    nCH := clientheight - 1;

    pen.mode  := pmCopy;
    pen.color := clBlack;

    { Outline }
    moveto (nCW, 0); lineto (nCW, nCH + 1);           { Right Side }
    moveto (0, nCH); lineto (nCW, nCH);               { Bottom }

    pen.color := clBtnFace;
    moveto (0, 0); lineto (0, nCH);                   { Left Side }
    moveto (0, 0); lineto (nCW, 0);                   { Top }
    moveto (2, 1); lineto (2, nCH - 1);               { Left Side }
    moveto (2, 2); lineto (nCW - 1, 2);               { Top }

    { Highlight / Shadow }
    pen.color := clWhite;
    moveto (1, 1); lineto (1, nCH - 1);               { Left Side }
    moveto (1, 1); lineto (nCW, 1);                   { Top }

    pen.color := clGray;
    moveto (nCW - 1, 2); lineto (nCW - 1, nCH);       { Right Side }
    moveto (2, nCH - 1); lineto (nCW - 1, nCH - 1);   { Bottom }
  end;
end;  { RU_DrawWindowBorder }


{-----------------------------------------------------------------------------
   Function : RU_IsPointInFakeCaption

   Returns  : True if point (x, y) is in the caption rect, false otherwise.
 -----------------------------------------------------------------------------}
function TRollupForm.RU_IsPointInFakeCaption (x, y: integer): boolean;
begin
  result := PtInRect (rcTitleBar, Point (x, y));

end;   { RU_IsPointInFakeCaption }


end.  { Unit: Rollups }

