(******************************************************************************
**
**
**  Object in this unit was originally written by and copyright
**                     by Andy Strong - Compuserve ID   : 100716,3015
**                                      Internet Address: andrews@nbaqsl.co.uk
**
**  Released into the public domain 14/02/97
**  TWINEXEDLLVER
**
**  Object for interrogating any DLL or Windows EXE which has a VERSION
**  resource embedded in it.  Either create it with the file name you want
**  to interrogate, or set the FileName property to the file name required.
**  Use the READ method to return the value requested - see TVersionTypes
**  ALL return values are strings.
**  For more information on Version Resources, reference TVS_FIXEDFILEINFO
**  and VERSIONINFO (3.1) in the Delphi on-line-help.
**
**  Note: Once a Version Resource is read in for a particular filename, it is
**        held in memory to allow any number of Member reads.  It will only
**        get cleared down when the FileName property is changed or the Clear
**        method is called.
**
**  Controls:
**  This code may be used, modified, included in applications without any
**  license agreements as long as the disclaimers are accepted, and the
**  comments are left intact ( Some Hope! <g> ),
**
**  Disclaimer:
**  This software is released into the public domain on the strict understanding
**  that neither myself nor any associates or companies I work for have any
**  liability explictly or implied.
**
**
*******************************************************************************)
unit Wexe_ver;

interface

uses
  Classes, Ver;

const
  MaxStrLength = 50;                        { Maximum needed string length }

type
  TVersionType = (vtComments, vtCompanyName, vtFileDescription, vtFileVersion,
                  vtInternalName, vtLegalCopyright, vtLegalTrademarks,
                  vtOriginalFilename, vtPrivateBuild, vtProductName,
                  vtProductVersion, vtSpecialBuild, vtSignature, vtStructVersion,
                  vtFileVersionInt, vtProductVersionInt, vtFileOS, vtFileType,
                  vtFileSubType, vtFileDate, vtFileFlags);

  TWinExeDllVer = class
  private
    { Storage areas }
    FFullFilename: string;                  { Full Drive/Path/and filename }
    FVersionExists: boolean;                { False if EXE/DLL has no Version info }
    FVersionBuffer: Pointer;                { Buffer containing Version info }
    FVersionSize: LongInt;                  { Buffer Size }
    FVersionHandle : LongInt;               { Version functions Handle }
    FFixedInfoBuffer: PVS_FIXEDFILEINFO;    { Fixed File Information block }
    FLanguagePath: array[0..79] of char;    { Language Path }
    FLanguageName: string;                  { Language Name }
    { Methods }
    function GetVersionInfo(Filename: openstring): boolean;

    procedure ClearBuffer;
    procedure SetFullFileName(Value: openstring);

  public
    constructor Create;
    constructor CreateWithFile(FullFile: openstring);
    destructor Destroy; override;

    function Read(VersionType: TVersionType): string;
    procedure Clear;

    property Filename: string read FFullFilename write SetFullFilename;
    property Language: string read FLanguageName;
    property VersionExists: boolean read FVersionExists;

  end;

implementation

uses
  SysUtils, WinProcs, WinTypes;

(*
**  Normal Constructor
*)
constructor TWinExeDllVer.Create;
begin
  inherited Create;                         { Create underlying object }
  ClearBuffer;                              { Reset all member variables }
end;


(*
**  Constructor to enable the object to be created with the desired File
*)
constructor TWinExeDllVer.CreateWithFile(FullFile: openstring);
begin
  inherited Create;                         { Create underlying object }
  ClearBuffer;                              { Reset all member variables }
  SetFullFilename(FullFile);                { Find and Read in File Version }
end;

(*
**  Overridden Destructor to release the resources used before destruction
*)
destructor TWinExeDllVer.Destroy;
begin
  ClearBuffer;                              { Release version resource memory }
  inherited Destroy;                        { Destroy the object }
end;

(*
**  Routine to release a Version resource buffer memory, and clear out
**  all private member variables
*)
procedure TWinExeDllVer.ClearBuffer;
begin
  if Assigned(FVersionBuffer) then
    FreeMem(FVersionBuffer, FVersionSize + 1);  { Release Version resource memory }
  FVersionBuffer := nil;
  FFixedInfoBuffer := nil;
  FFullFilename := '';
  FVersionExists := False;
  FVersionSize := 0;
  FVersionHandle := 0;
  FLanguageName := '';
  FillChar(FLanguagePath, SizeOf(FLanguagePath), 0);
end;

(*
**  Procedure to clear down any memory, and the currently loaded Version Resource
*)
procedure TWinExeDllVer.Clear;
begin
  ClearBuffer;                              { Reset all member variables }
end;

(*
**  Procedure to attempt to read the Version resource from the passed filename
*)
procedure TWinExeDllVer.SetFullFilename(Value: openstring);
begin
  if Value <> FFullFileName then            { If a new file name }
  begin
    if Value = '' then
      ClearBuffer                           { Release old information }
    else
      FVersionExists :=
        GetVersionInfo(Value);              { Load up new Version info }
    FFullFilename := Value;                 { Set internal file name }
  end;
end;

(*
**  Routine to actually read in the Version Resource information (if there is
**  one) into the objects control.
*)
function TWinExeDllVer.GetVersionInfo(Filename: openstring): boolean;
var
  PCharFile: PChar;
  PLangName: PChar;
  Len: word;
  LangCharSetIDArray: array[1..2] of word;  { Language ID }
  TranslationInfoPointer: Pointer;          { Language set pointer }
begin
  Result := False;                          { Assume failure }
  ClearBuffer;                              { Reset all member variables }
  PLangName := StrAlloc(MaxStrLength);      { Allocate space for Language Name }
  PCharFile :=
    StrAlloc(Length(Filename) + 1);         { Allocate Filename string }
  StrPCopy(PCharFile, Filename);            { Copy Pascal string to PCHAR }

  try
    FVersionSize := GetFileVersionInfoSize(
      PCharFile, FVersionHandle);           { Attempt to read Version Info }
    if FVersionSize > 0 then                { If we got one }
    begin
      GetMem(FVersionBuffer,
        FVersionSize + 1);                  { Allocate Memory for it }
      FillChar(FVersionBuffer^,
        FVersionSize + 1, 0);               { Zero memory allocated }
      try
        if not GetFileVersionInfo(PCharFile,
          FVersionHandle, FVersionSize,
          FVersionBuffer) then              { If we can't read Version info }
          ClearBuffer;                      { Clear internals out }

        if VerQueryValue(FVersionBuffer,
          '\VarFileInfo\Translation',
          TranslationInfoPointer, Len) then { Get Language Translation code }
        begin
          LangCharSetIDArray[1] := LoWord(
            PLongint(TranslationInfoPointer)^);
          LangCharSetIDArray[2] := HiWord(
            PLongint(TranslationInfoPointer)^);{ Set Language LongInt codes }
          wvSprintf(FLanguagePath, '\StringFileInfo\%04x%04x\',
            LangCharSetIDArray);            { Slot codes into Main Path String }
          VerLanguageName(LangCharSetIDArray[1],
            PLangName, MaxStrLength);       { Retrieve Language desc text }
          FLanguageName := StrPas(PLangName);{ Copy to Pascal String }

          if VerQueryValue(FVersionBuffer, '\',
            Pointer(FFixedInfoBuffer), Len) then
            Result := True;                 { Get Pointer to Fixed File info }

        end else
          raise Exception.Create('Error retrieving Language information');
      except
        ClearBuffer;
        raise;
      end;
    end;
  finally
    StrDispose(PLangName);                  { Release PCHAR strings memory }
    StrDispose(PCharFile);
  end;
end;

(*
**  General purpose Version Information Member read function
*)
function TWinExeDllVer.Read(VersionType: TVersionType): string;

  (*
  **  Reads the relevant Member information given the path string to it
  **  (SearchStr) and the Member information required (StrName)
  *)
  function ReadStrFileInfo(const StrName: openstring): string;
  var
    SearchStr: PChar;
    VersPChar: Pointer;
    VerSelection: PChar;
    Len: word;
  begin
    Result := '';                           { Assume no return string }
    SearchStr := StrAlloc(MaxStrLength);    { Allocate Memory for Search Path }
    FillChar(SearchStr^, SizeOf(SearchStr), 0);
    VerSelection := StrAlloc(MaxStrLength); { Allocate space Member Text }
    try
      StrPCopy(VerSelection, StrName);      { Copy Member Descr to PCHAR }
      StrCopy(SearchStr, FLanguagePath);    { Copy base path to PCHAR }
      StrCat(SearchStr, VerSelection);      { Add member Desc text to base path }
      if VerQueryValue(FVersionBuffer,
        SearchStr, VersPChar, Len) then     { If Member information found }
        Result := StrPas(VersPChar);        { Return as a string }
    finally
      StrDispose(VerSelection);             { Release PCHAR memory }
      StrDispose(SearchStr);
    end
  end;

  (*
  **  Function to make a Double LongInt into a version string of the form:
  **    0.0.2.1
  *)
  function MakeFullVersion(MS, LS: Longint): string;
  begin
    Result := IntToStr((MS and $FFFF0000) shr $10) + '.' +
              IntToStr( MS and $FFFF) + '.' +
              IntToStr((LS and $FFFF0000) shr $10) + '.' +
              IntToStr( LS and $FFFF);
  end;

var
  VersPCharPath: PChar;
  TmpLong: LongInt;
begin
  Result := '';                             { Assume nothing }
  if FVersionExists then
  begin
    try
      case VersionType of
        vtComments:
          Result := ReadStrFileInfo('Comments');
        vtCompanyName:
          Result := ReadStrFileInfo('CompanyName');
        vtFileDescription:
          Result := ReadStrFileInfo('FileDescription');
        vtFileVersion:
          Result := ReadStrFileInfo('FileVersion');
        vtInternalName:
          Result := ReadStrFileInfo('InternalName');
        vtLegalCopyright:
          Result := ReadStrFileInfo('LegalCopyright');
        vtLegalTrademarks:
          Result := ReadStrFileInfo('LegalTrademarks');
        vtOriginalFilename:
          Result := ReadStrFileInfo('OriginalFilename');
        vtPrivateBuild:
          Result := ReadStrFileInfo('PrivateBuild');
        vtProductName:
          Result := ReadStrFileInfo('ProductName');
        vtProductVersion:
          Result := ReadStrFileInfo('ProductVersion');
        vtSpecialBuild:
          Result := ReadStrFileInfo('SpecialBuild');
        vtSignature:
          Result := Format('%8x', [FFixedInfoBuffer^.dwSignature]);
        vtStructVersion:
          Result := IntToStr((FFixedInfoBuffer^.dwStrucVersion and $FFFF0000)
                    shr $10) + '.' +
                    IntToStr( FFixedInfoBuffer^.dwStrucVersion and $FFFF);
        vtFileVersionInt:
          Result := MakeFullVersion(FFixedInfoBuffer^.dwFileVersionMS,
                                    FFixedInfoBuffer^.dwFileVersionLS);
        vtProductVersionInt:
          Result := MakeFullVersion(FFixedInfoBuffer^.dwProductVersionMS,
                                    FFixedInfoBuffer^.dwProductVersionLS);
        vtFileOS:
          Result := IntToStr(FFixedInfoBuffer^.dwFileOS);
        vtFileType:
          Result := IntToStr(FFixedInfoBuffer^.dwFileType);
        vtFileSubType:
          Result := IntToStr(FFixedInfoBuffer^.dwFileSubType);
        vtFileDate:
          Result := IntToStr(FFixedInfoBuffer^.dwFileDateMS) + '.' +
                    IntToStr(FFixedInfoBuffer^.dwFileDateLS);
        vtFileFlags:
          Result := IntToStr(FFixedInfoBuffer^.dwFileFlags and
                             FFixedInfoBuffer^.dwFileFlagsMask);
      end;
    except
      raise Exception.Create('Error retrieving Version information');
    end;
  end;
end;

end.
